# waddle_logger.py

import psutil
import platform
import argparse
import json
import sys
import os
import subprocess
import threading
import time
from datetime import datetime
from pynvml import *
from typing import Any, Dict, Optional
import uuid

class WaddleLogger:
    def __init__(self, db_root, project, id=None, config=None, use_gpu_metrics=True, server_url=None):
        self.db_root = db_root
        self.project = project
        self.id = id or datetime.now().strftime('%Y%m%d_%H%M%S')
        self.log_folder = os.path.join(db_root, self.project, self.id)
        os.makedirs(self.log_folder, exist_ok=True)
        self.config: argparse.Namespace = config
        self.use_gpu_metrics = use_gpu_metrics
        self.server_url = server_url

        # Log initial system, CLI parameters, and code
        self.log_run_info()

    def _get_file_prefix(self):
        return f"{int(time.time() * 1000)}_{uuid.uuid4().hex}"

    def log_run_info(self):
        # Get system information
        python_version = sys.version
        os_info = platform.platform()
        cpu_info = platform.processor()
        total_memory = psutil.virtual_memory().total / (1024 ** 3)  # Convert to GB

        # Retrieve Git commit hash if available
        git_hash = self.get_git_commit_hash()

        # Read the running script (if accessible)
        code_data = self.get_running_script_code()

        # Prepare the CLI parameters as JSON
        cli_params_json = json.dumps(sys.argv)

        # Prepare the GPU info
        gpu_info = json.dumps(self.get_gpu_system_metrics())

        # Prepare the run info dictionary
        run_info = {
            "run_info": {
                'id': self.id,
                'start_time': datetime.now().isoformat(),
                'cli_params': cli_params_json,
                'python_version': python_version,
                'os_info': os_info,
                'cpu_info': cpu_info,
                'total_memory': total_memory,
                'gpu_info': gpu_info,
                'code': code_data.decode('utf-8'),
                'git_hash': git_hash,
                'timestamp': datetime.now().isoformat()
            }
        }

        # Write run_info to a file in the log folder
        run_info_file = os.path.join(self.log_folder, f'{self._get_file_prefix()}.run_info.json')
        with open(run_info_file, 'w') as f:
            json.dump(run_info, f)

    def get_git_commit_hash(self):
        """
        Retrieves the current Git commit hash if the script is in a Git repository.
        """
        try:
            git_hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode()
        except Exception:
            git_hash = None
        return git_hash

    def get_running_script_code(self):
        """
        Reads and returns the content of the running script.
        """
        try:
            with open(__file__, 'r') as f:
                return f.read().encode('utf-8')  # Encode to binary for storage
        except Exception as e:
            print(f"Could not read the script: {e}")
            return b''  # Return an empty byte if reading fails

    def get_gpu_system_metrics(self):
        """
        Collects system information about GPU utilization, temperature, etc.
        """
        gpu_metrics = []
        if self.use_gpu_metrics:
            try:
                gpu_count = nvmlDeviceGetCount()
                for i in range(gpu_count):
                    handle = nvmlDeviceGetHandleByIndex(i)
                    gpu_metrics.append({
                        'gpu_index': i,
                        'name': nvmlDeviceGetName(handle),
                        'temperature': nvmlDeviceGetTemperature(handle, NVML_TEMPERATURE_GPU),
                        'memory_used': nvmlDeviceGetMemoryInfo(handle).used / (1024 ** 3),  # GB
                        'memory_total': nvmlDeviceGetMemoryInfo(handle).total / (1024 ** 3),  # GB
                        'utilization': nvmlDeviceGetUtilizationRates(handle).gpu
                    })
            except Exception as e:
                print(f"Error getting GPU metrics: {e}")
        return gpu_metrics

    def log(self, data: Dict[str, Any], step: Optional[int], category='default', timestamp=None):
        """
        Writes log entries as JSON files to the log folder (solo mode) or sends via REST API (distributed mode).
        """
        timestamp = timestamp or datetime.now().isoformat()
        for name, value in data.items():
            log_entry = {
                'id': self.id,
                'project': self.project,
                'step': step or 0,
                'category': category,
                'name': name,
                'value': value,
                'timestamp': timestamp
            }
            # Write to local folder
            filename = f"{self._get_file_prefix()}.json"
            temp_filename = f"{filename}.tmp"
            filepath = os.path.join(self.log_folder, temp_filename)
            # Write to a temporary file
            with open(filepath, 'w') as f:
                json.dump(log_entry, f)
            # Rename to the final filename to ensure atomicity
            final_filepath = os.path.join(self.log_folder, filename)
            os.rename(filepath, final_filepath)

    def log_gpu_metrics_periodically(self, interval=60):
        """
        Periodically logs GPU system metrics every `interval` seconds in a separate thread.
        """
        def log_metrics():
            while True:
                gpu_metrics_list = self.get_gpu_system_metrics()
                for metric in gpu_metrics_list:
                    data = {}
                    for key, value in metric.items():
                        if key != 'gpu_index' and key != 'name':  # Skip index and name in logs
                            data[f'gpu_{metric["gpu_index"]}_{key}'] = value
                    self.log(data=data, step=None, category='gpu_system', timestamp=datetime.now().isoformat())
                time.sleep(interval)

        # Create and start a thread for logging GPU metrics
        logging_thread = threading.Thread(target=log_metrics, daemon=True)
        logging_thread.start()

# Global variables and functions
config: argparse.Namespace = argparse.Namespace()
run: WaddleLogger = None
server_process: Optional[subprocess.Popen] = None

def _assign_config(app_config):
    global config
    config = app_config

def init(project, log_root='.waddle/logs', config=None, use_gpu_metrics=True, gpu_metrics_interval=60, server_url=None, server_port=8000, server_bind="127.0.0.1"):
    global run
    global server_process

    os.makedirs(log_root, exist_ok=True)

    # Assign the passed config to the global config
    config = config or argparse.Namespace()
    _assign_config(config)

    if use_gpu_metrics:
        # Check if the pynvml package is available
        try:
            nvmlInit()
        except Exception as e:
            print(f"Could not initialize NVML: {e}")
            use_gpu_metrics = False

    if server_url is None:
        # Start the waddle server as a subprocess
        server_cmd = ['waddle', '--mode', 'server', '--server-port', str(server_port), '--server-bind', server_bind, '--db-root', log_root, '--project', project, '--watch-folder', os.path.join(log_root, project)]
        server_process = subprocess.Popen(server_cmd)
        # Allow the server some time to start
        time.sleep(2)
        server_url = 'http://localhost:8000'
    else:
        if not server_url:
            raise ValueError("In distributed mode, server_url must be specified.")

    run = WaddleLogger(db_root=log_root, project=project, use_gpu_metrics=use_gpu_metrics, config=config, server_url=server_url)

    print("Waddle Logger initialized.")
    print("Run ID:", run.id)

    if use_gpu_metrics and gpu_metrics_interval > 0:
        run.log_gpu_metrics_periodically(interval=gpu_metrics_interval)

    return run

def finish():
    global server_process
    if server_process is not None:
        server_process.terminate()
        server_process.wait()
        server_process = None

def log(category, data, step, timestamp=None):
    if run is None:
        raise ValueError("WaddleLogger is not initialized. Please call `init()` first.")
    if not isinstance(data, dict):
        raise ValueError("The data must be a dictionary.")
    run.log(data=data, step=step, category=category, timestamp=timestamp)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--project-name', type=str, default='experiment')
    parser.add_argument('--mode', type=str, choices=['solo', 'distributed'], default='solo')
    parser.add_argument('--server-url', type=str, default=None)
    args = parser.parse_args()

    # Initialize WaddleLogger
    init(project=args.project_name, config=args, mode=args.mode, server_url=args.server_url)

    # Simulate the rest of your ML code here
    # The GPU logging will run in the background
    try:
        for step in range(10):
            log_data = {'loss': 0.01 * step}
            log(category='model', data=log_data, step=step)
            time.sleep(5)  # Simulating training steps
    finally:
        # Finish up and terminate the server process
        finish()

    print("Done!")



import argparse
import asyncio
import glob
import json
import os
import threading
import time
from contextlib import asynccontextmanager
import logging

import duckdb
import uvicorn
from fastapi import FastAPI, HTTPException, WebSocket, WebSocketDisconnect, Request
from fastapi.responses import HTMLResponse, FileResponse
from fastapi.templating import Jinja2Templates
import pandas as pd
from datetime import datetime

logger = logging.getLogger(__name__)

waddle_server_instance = None

@asynccontextmanager
async def lifespan(app: FastAPI):
    global waddle_server_instance
    parser = argparse.ArgumentParser()
    parser.add_argument('--db-root', type=str, default='.waddle')
    parser.add_argument('--project', type=str, default='experiment')
    parser.add_argument('--watch-folder', type=str, default=None)
    args, _ = parser.parse_known_args()

    loop = asyncio.get_running_loop()
    waddle_server_instance = WaddleServer(db_root=args.db_root, project=args.project, watch_folder=args.watch_folder, loop=loop)
    yield
    if waddle_server_instance:
        waddle_server_instance.stop()

app = FastAPI(lifespan=lifespan)

templates = Jinja2Templates(directory=os.path.join(os.path.dirname(__file__), "templates"))

class ConnectionManager:
    def __init__(self):
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)
        logger.info(f"WebSocket connection established: {websocket.client}")

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)
        logger.info(f"WebSocket connection closed: {websocket.client}")

    async def broadcast(self, message: dict):
        for connection in self.active_connections:
            try:
                await connection.send_json(message)
            except Exception as e:
                logger.error(f"Error sending message to client {connection.client}: {e}")
                self.disconnect(connection)

manager = ConnectionManager()

class WaddleServer:
    def __init__(self, db_root, project, watch_folder=None, loop=None, watch_delay=10):
        self.db_root = db_root
        self.project = project
        self.db_path = os.path.join(db_root, f"{project}.db")
        self.watch_folder = watch_folder
        self.loop = loop
        self.watch_delay = watch_delay

        self.con = duckdb.connect(self.db_path)

        # add project info to the database
        self.con.execute('''
            CREATE TABLE IF NOT EXISTS project_info (
                name VARCHAR,
                description VARCHAR,
                timestamp TIMESTAMP,
                PRIMARY KEY (name)
            );
        ''')
        # non-destructively add project info
        self.con.execute('''
            INSERT OR IGNORE INTO project_info VALUES (?, ?, ?)
        ''', (project, None, datetime.now().isoformat()))

        # Create tables if they don't exist
        self.con.execute('''
            CREATE TABLE IF NOT EXISTS run_info (
                id VARCHAR,
                start_time TIMESTAMP,
                cli_params JSON,
                python_version VARCHAR,
                os_info VARCHAR,
                cpu_info VARCHAR,
                total_memory DOUBLE,
                gpu_info JSON,
                code BLOB,
                git_hash VARCHAR,
                timestamp TIMESTAMP,
                PRIMARY KEY (id)
            );
        ''')

        self.con.execute('''
            CREATE TABLE IF NOT EXISTS logs (
                id VARCHAR,
                step INTEGER,
                category VARCHAR,
                name VARCHAR,
                value_double DOUBLE,
                value_string VARCHAR,
                value_blob BLOB,
                timestamp TIMESTAMP
            );
        ''')

        if self.watch_folder:
            # Start the folder-watching thread
            self.watching = True
            self.watch_thread = threading.Thread(target=self.watch_folder_for_logs, daemon=True)
            self.watch_thread.start()

    def stop(self):
        if self.watch_folder:
            self.watching = False
            self.watch_thread.join()

    def watch_folder_for_logs(self):
        try:
            while self.watching:
                logger.info(f"{time.time()} Watching folder for logs...")

                # Get all directories recursively
                log_folders = glob.glob(os.path.join(self.watch_folder, '**'), recursive=True)
                log_folders = [f for f in log_folders if os.path.isdir(f)]

                for log_folder in log_folders:
                    logger.info(f"Processing log folder: {log_folder}")

                    # Process log entries recursively
                    log_files = glob.glob(os.path.join(log_folder, '**', '*.json'), recursive=True)
                    for log_file in log_files:
                        try:
                            logging.info(f"Processing log file: {log_file}")
                            with open(log_file, 'r') as f:
                                log_entry = json.load(f)
                                # Ingest log entry
                                self.ingest_log_entry(log_entry)
                            # Delete the log file after processing
                            os.remove(log_file)
                        except Exception as e:
                            logger.error(f"Error processing log file {log_file}: {e}")

                time.sleep(self.watch_delay)
        except Exception as e:
            logger.error(f"Error in watch_folder_for_logs: {e}")

    def ingest_run_info(self, run_info):
        run_id = run_info['id']
        self.con.execute("SELECT 1 FROM run_info WHERE id = ?", (run_id,))
        if not self.con.fetchall():
            # Insert into run_info table
            self.con.execute('''
                INSERT INTO run_info VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            ''', (
                run_info['id'],
                run_info['start_time'],
                run_info['cli_params'],
                run_info['python_version'],
                run_info['os_info'],
                run_info['cpu_info'],
                run_info['total_memory'],
                run_info['gpu_info'],
                run_info['code'].encode('utf-8'),
                run_info['git_hash'],
                run_info['timestamp']
            ))

    def ingest_log_entry(self, log_entry):
        try:
            # Handle run_info separately
            if 'run_info' in log_entry:
                self.ingest_run_info(log_entry['run_info'])
                return

            # Determine the type of the value and insert accordingly
            value = log_entry['value']
            value_double = None
            value_string = None
            value_blob = None

            if isinstance(value, (int, float)):
                value_double = value
            elif isinstance(value, str):
                value_string = value
            elif isinstance(value, (list, dict)):
                # Convert lists and dicts to JSON string
                value_string = json.dumps(value)
            elif isinstance(value, bytes):
                value_blob = value
            else:
                # For other types, store as string
                value_string = json.dumps(value)

            self.con.execute('''
                INSERT INTO logs (
                    id, step, category, name, value_double, value_string, value_blob, timestamp
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
            ''', (
                log_entry['id'], log_entry['step'], log_entry['category'], log_entry['name'],
                value_double, value_string, value_blob, log_entry['timestamp']
            ))

            # Broadcast the new log entry to connected WebSocket clients
            asyncio.run_coroutine_threadsafe(manager.broadcast(log_entry), self.loop)

        except Exception as e:
            logger.error(f"Error ingesting log entry: {e}")
            raise

@app.post("/ingest")
async def ingest_log(log_entry: dict):
    if not waddle_server_instance:
        raise HTTPException(status_code=500, detail="Server not initialized")
    try:
        waddle_server_instance.ingest_log_entry(log_entry)
        return {"status": "success"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


def sanitize_df(df: pd.DataFrame) -> pd.DataFrame:
    # coalesce the value_double and value_string columns into a value column
    # df['value'] = df['value_double'].combine_first(df['value_string'])

    for i, row in df.iterrows():
        if not pd.isna(row['value_double']):
            df.at[i, 'value'] = row['value_double']
        else:
            df.at[i, 'value'] = row['value_string']

    # drop the value_double and value_string columns
    df = df.drop(columns=['value_double', 'value_string'])

    # Convert timestamps to ISO format
    if 'timestamp' in df.columns:
        df['timestamp'] = pd.to_datetime(df['timestamp'], errors='coerce').dt.strftime('%Y-%m-%dT%H:%M:%S')

    df = df.sort_values(by='step')

    return df


@app.get("/info")
async def get_info():
    if not waddle_server_instance:
        raise HTTPException(status_code=500, detail="Server not initialized")
    try:
        df = waddle_server_instance.con.execute("SELECT * FROM project_info").fetchdf()
        return df.to_dict(orient='records')
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/runs")
async def get_runs():
    if not waddle_server_instance:
        raise HTTPException(status_code=500, detail="Server not initialized")
    try:
        df = waddle_server_instance.con.execute("SELECT id,start_time FROM run_info ORDER BY start_time DESC").fetchdf()
        df['start_time'] = pd.to_datetime(df['start_time'], errors='coerce').dt.strftime('%Y-%m-%dT%H:%M:%S')
        return df.to_dict(orient='records')
    except Exception as e:
        import traceback
        print(e)
        logger.error(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/run/{run_id}")
async def get_run(run_id: str, history: int = 10):
    if not waddle_server_instance:
        raise HTTPException(status_code=500, detail="Server not initialized")
    try:
        # first get the list of names
        df = waddle_server_instance.con.execute("SELECT DISTINCT name FROM logs WHERE id = ?", (run_id,)).fetchdf()
        # for each name then get the history of that name and collect it in a list
        data = []
        for name in df['name']:
            df = waddle_server_instance.con.execute("SELECT id,step,category,name,value_double,value_string,timestamp FROM logs WHERE id = ? AND name = ? ORDER BY step DESC LIMIT ?", (run_id,name,history,)).fetchdf()
            clean_df = sanitize_df(df).to_dict(orient='records')
            data.extend(clean_df)
        logger.info(f"Returning {len(data)} records for run {run_id}")
        return data
    except Exception as e:
        import traceback
        print(e)
        logger.error(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))

# @app.get("/data")
# async def get_data(history: int = 2000):
#     history = min(max(history, 1), 2000)  # Limit history to between 1 and 1000
#     if not waddle_server_instance:
#         raise HTTPException(status_code=500, detail="Server not initialized")
#     try:
#         df = waddle_server_instance.con.execute("SELECT * FROM logs").fetchdf()
#         df = sanitize_df(df)
#         df = df[-history:]
#         x = df.to_dict(orient='records')
#         return x
#     except Exception as e:
#         import traceback
#         logger.error(traceback.format_exc())
#         raise HTTPException(status_code=500, detail=str(e))

@app.get("/", response_class=HTMLResponse)
async def read_root(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})

# get static assets
@app.get("/static/{filename}")
async def get_static(filename: str):
    filename = os.path.basename(filename)
    filepath = os.path.join(os.path.dirname(__file__), "static", filename)
    return FileResponse(filepath)


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()  # Keep the connection open
    except WebSocketDisconnect:
        manager.disconnect(websocket)

def main(port=8000, bind="127.0.0.1", log_level="critical"):
    logging.getLogger("uvicorn.error").handlers = []
    logging.getLogger("uvicorn.error").propagate = False

    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f'Invalid log level: {log_level}')
    logging.basicConfig(level=numeric_level)
    logger.setLevel(numeric_level)

    logger.info("Starting Waddle server.")
    print("Starting Waddle server.")
    logger.info(f"Listening at http://{bind}:{port}")
    print(f"Listening at http://{bind}:{port}")
    uvicorn.run("waddle.server:app", host=bind, port=port, log_level=log_level, lifespan="on")
    logger.info("Waddle server stopped.")
    print("Waddle server stopped.")

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--server-port", type=int, default=8000)
    parser.add_argument("--server-bind", type=str, default="127.0.0.1")
    parser.add_argument("--log-level", type=str, default="critical")
    args = parser.parse_args()
    main(port=args.server_port, bind=args.server_bind, log_level=args.log_level)



import os
from setuptools import setup, find_packages

directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

# use the requirements.txt to install dependencies
with open(os.path.join(directory, 'requirements.txt')) as f:
    required_packages = f.read().splitlines()

setup(
    name='waddleml',
    packages=find_packages(),
    version='0.1.3',
    description='WaddleML is a machine learning stats tracker built on DuckDB',
    author='Brian Guarraci',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=required_packages,
    python_requires='<=3.12,>=3.8',
    include_package_data=True,
    zip_safe=False,
    test_suite='tests',
    scripts=['scripts/waddle']
)
